import time

import cv2
import numpy as np
import PIL

from ... import rosys
from ...geometry import Point
from ..image import ImageSize


class SimulatedDevice:

    def __init__(self,
                 id: str,  # pylint: disable=redefined-builtin
                 size: ImageSize,
                 color: str = '#ffffff') -> None:
        self.id = id
        self.size = size
        self.color = color
        self.creation_time: float = rosys.time()

    def create_image_data(self) -> bytes:
        img = PIL.Image.new('RGB', size=(self.size.width, self.size.height), color=self.color)
        d = PIL.ImageDraw.Draw(img)
        text = f'{self.id}: {time.time():.2f}'
        position = floating_text_position(self.size.width, self.size.height)

        d.text((position.x, position.y), text, fill=(0, 0, 0))
        d.text((position.x + 1, position.y + 1), text, fill=(255, 255, 255))

        _, encoded_image = cv2.imencode('.jpg', np.array(img)[:, :, ::-1])  # NOTE: cv2 expects BGR
        return encoded_image.tobytes()


def floating_text_position(box_width: int, box_height: int, speed: float = 100, angle: float = np.deg2rad(45)) -> Point:
    """Calculate the position of the text that floats around the image."""
    t = time.time()
    return Point(x=(speed * t * np.cos(angle)) % box_width,
                 y=(speed * t * np.sin(angle)) % box_height)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from .line import Line
from .point import Point


@dataclass(slots=True, kw_only=True)
class LineSegment:
    point1: Point
    point2: Point

    @property
    def direction(self) -> float:
        return self.point1.direction(self.point2)

    @property
    def line(self) -> Line:
        return Line.from_points(self.point1, self.point2)

    def intersect(self, other: LineSegment, allow_outside=False) -> Optional[Point]:
        # https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection#Given_two_points_on_each_line_segment
        p1 = self.point1
        p2 = self.point2
        p3 = other.point1
        p4 = other.point2
        d = (p1.x - p2.x) * (p3.y - p4.y) - (p1.y - p2.y) * (p3.x - p4.x)
        s = ((p1.x - p3.x) * (p3.y - p4.y) - (p1.y - p3.y) * (p3.x - p4.x)) / d
        t = ((p2.x - p1.x) * (p1.y - p3.y) - (p2.y - p1.y) * (p1.x - p3.x)) / d
        if allow_outside or (0 <= s <= 1 and 0 <= t <= 1):
            return Point(x=p1.x + s * (p2.x - p1.x), y=p1.y + s * (p2.y - p1.y))
        return None

    def distance(self, point: Point) -> float:
        # https://stackoverflow.com/a/6853926/3419103
        a = point.x - self.point1.x
        b = point.y - self.point1.y
        c = self.point2.x - self.point1.x
        d = self.point2.y - self.point1.y
        dot = a * c + b * d
        len_sq = c * c + d * d
        t = dot / len_sq if len_sq > 0 else -1
        p = self.point1 if t < 0 else self.point2 if t > 1 else Point(x=self.point1.x + t * c, y=self.point1.y + t * d)
        return p.distance(point)



from abc import ABC, abstractmethod
from itertools import groupby
from typing import Sequence

import humanize
from matplotlib import colormaps
from matplotlib.colors import to_hex
from nicegui import ui

from .kpi_buckets import Month, TimeBucket, Week
from .kpi_chart import KpiChart
from .kpi_logger import KpiLogger, str_to_date


class kpi_page(ABC):

    def __init__(self, kpi_logger: KpiLogger) -> None:
        @ui.page('/kpis')
        def page():
            if self.language == 'de':
                humanize.activate('de')
            else:
                humanize.deactivate()
            with ui.row().style('margin:1em'):
                ui.markdown(f'### {self.title}').style('margin:1.5em;margin-top:-1.2em;')
                toggle = ui.toggle(self.timespans, value=list(self.timespans)[0], on_change=lambda e: show(e.value))
            with ui.row().classes('w-full'):
                ui_charts = []
                for chart in self.charts:
                    ui_charts.append(ui.echart({
                        'title': {'text': chart.title},
                        'xAxis': {'type': 'category'},
                        'yAxis': {'type': 'value', 'name': chart.unit},
                        'legend': {'bottom': 10},
                        'tooltip': {},
                        'color': [to_hex(colormaps[chart.colormap or 'viridis'](i / 10)) for i in range(9, -1, -2)],
                        'series': [],
                    }))
            ui.timer(5, lambda: show(toggle.value))

            def show(num_days: int) -> None:
                time_buckets: Sequence[TimeBucket]
                if num_days <= 7:
                    time_buckets = kpi_logger.days[-num_days:]
                elif num_days % 7 == 0:
                    weeks = [list(v) for _, v in groupby(kpi_logger.days,
                                                         key=lambda d: str_to_date(d.date).strftime('%y-%W'))][-num_days//7:]
                    time_buckets = [Week.from_buckets(days_in_week) for days_in_week in weeks]
                elif num_days % 30 == 0 and num_days <= 90:
                    months = [list(v) for _, v in groupby(kpi_logger.days,
                                                          key=lambda d: str_to_date(d.date).strftime('%y-%m'))][-num_days//30:]
                    time_buckets = [Month.from_buckets(days_in_month) for days_in_month in months]
                else:
                    raise ValueError(f'Unsupported number of days: {num_days}')

                for chart, ui_chart in zip(self.charts, ui_charts):
                    keys = set(key for day in time_buckets for key in day.incidents if key in chart.indicators)
                    data = {chart.indicators[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
                    styling = {'type': 'bar', 'stack': 'total', 'emphasis': {'focus': 'series'}}
                    ui_chart.options['xAxis']['data'] = [_label(b) for b in time_buckets]
                    ui_chart.options['series'] = [
                        {**styling, 'name': k, 'data': [round(item * chart.scale, 2) for item in v]}
                        for k, v in sorted(data.items())
                    ]
                    ui_chart.update()

    @property
    def language(self) -> str:
        return 'en'

    @property
    @abstractmethod
    def title(self) -> str:
        pass

    @property
    @abstractmethod
    def timespans(self) -> dict[int, str]:
        pass

    @property
    @abstractmethod
    def charts(self) -> list[KpiChart]:
        pass


def _label(time_bucket: TimeBucket) -> str:
    if isinstance(time_bucket, Month):
        return str_to_date(time_bucket.date).strftime(r'%b')
    if isinstance(time_bucket, Week):
        return f'#{str_to_date(time_bucket.date).isocalendar()[1]}'
    return humanize.naturaldate(str_to_date(time_bucket.date))



from __future__ import annotations

import time
from typing import Optional, overload

import cv2
import numpy as np
from scipy import ndimage

from .area import Area
from .binary_renderer import BinaryRenderer
from .grid import Grid
from .obstacle import Obstacle
from .robot_renderer import RobotRenderer


class ObstacleMap:

    def __init__(self, grid, map_, robot_renderer, deadline=None) -> None:
        self.grid = grid
        self.map = map_
        self.stack = np.zeros(grid.size, dtype=bool)
        self.dist_stack = np.zeros(self.stack.shape)
        for layer in range(grid.size[2]):
            _, _, yaw = grid.from_3d_grid(0, 0, layer)
            kernel = robot_renderer.render(grid.pixel_size, yaw).astype(np.uint8)
            self.stack[:, :, layer] = cv2.dilate(self.map.astype(np.uint8), kernel)
            self.dist_stack[:, :, layer] = \
                ndimage.distance_transform_edt(~self.stack[:, :, layer]) * grid.pixel_size
            if deadline and time.time() > deadline:
                raise TimeoutError('obstacle map creation took too long')

        # NOTE: when yaw wraps around, map_coordinates should wrap around on axis 2
        self.stack = np.dstack((self.stack, self.stack[:, :, :1]))
        self.dist_stack = np.dstack((self.dist_stack, self.dist_stack[:, :, :1]))

    @staticmethod
    def from_list(grid, obstacles, robot_renderer) -> ObstacleMap:
        map_ = np.zeros(grid.size[:2], dtype=bool)
        for x, y, w, h in obstacles:
            r0, c0 = grid.to_grid(x, y)
            r1, c1 = grid.to_grid(x + w, y + h)
            map_[int(np.round(r0)):int(np.round(r1))+1,
                 int(np.round(c0)):int(np.round(c1))+1] = True
        return ObstacleMap(grid, map_, robot_renderer)

    @staticmethod
    def from_world(robot_outline: list[tuple[float, float]],
                   areas: list[Area],
                   obstacles: list[Obstacle],
                   grid: Grid,
                   deadline: Optional[float] = None) -> ObstacleMap:
        robot_renderer = RobotRenderer(robot_outline)
        has_areas = any(len(a.outline) > 2 for a in areas)
        binary_renderer = BinaryRenderer(grid.size[:2], fill_value=has_areas)
        for area in areas:
            binary_renderer.polygon(np.array([grid.to_grid(p.x, p.y)[::-1] for p in area.outline]), False)
            if deadline and time.time() > deadline:
                raise TimeoutError('obstacle map creation took too long')
        for obstacle in obstacles:
            binary_renderer.polygon(np.array([grid.to_grid(p.x, p.y)[::-1] for p in obstacle.outline]))
            if deadline and time.time() > deadline:
                raise TimeoutError('obstacle map creation took too long')
        return ObstacleMap(grid, binary_renderer.map, robot_renderer, deadline)

    def test(self, x, y, yaw):
        row, col, layer = self.grid.to_3d_grid(x, y, yaw)
        return ndimage.map_coordinates(self.stack, [[row], [col], [layer]], order=0)

    t_lookup = [np.linspace(0, 1, i) for i in range(360)]

    def _create_poses(self, spline, backward) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        @overload
        def pose(t: float) -> tuple[float, float, float]: ...

        @overload
        def pose(t: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]: ...

        def pose(t: float | np.ndarray) -> tuple[float | np.ndarray, float | np.ndarray, float | np.ndarray]:
            return (
                spline.x(t),
                spline.y(t),
                spline.yaw(t) + [0, np.pi][backward],
            )

        row0, col0, layer0 = self.grid.to_3d_grid(*pose(0.0))
        row1, col1, layer1 = self.grid.to_3d_grid(*pose(1.0))
        num_rows = int(abs(row1 - row0))
        num_cols = int(abs(col1 - col0))
        num_layers = int(abs(layer1 - layer0))
        n = max(num_rows, num_cols, num_layers)
        t = ObstacleMap.t_lookup[n] if n < len(ObstacleMap.t_lookup) else np.linspace(0, 1, n)
        return pose(t)

    def test_spline(self, spline, backward=False) -> bool:
        return self.test(*self._create_poses(spline, backward)).any()

    def get_distance(self, x, y, yaw) -> np.ndarray:
        row, col, layer = self.grid.to_3d_grid(x, y, yaw)
        return ndimage.map_coordinates(self.dist_stack, [[row], [col], [layer]], order=0)

    def get_minimum_spline_distance(self, spline, backward=False) -> float:
        return self.get_distance(*self._create_poses(spline, backward)).min()



import pytest

pytest.register_assert_rewrite('rosys.testing.helpers')  # nopep8

# pylint: disable=wrong-import-position
from . import log_configuration
from .helpers import approx, assert_point, assert_pose, automate_drive_to, forward



from __future__ import annotations

import numpy as np

from .binary_renderer import BinaryRenderer


class RobotRenderer:

    def __init__(self, outline) -> None:
        self.outline = outline
        self.rendered_outline: np.ndarray = np.zeros((0, 2))

    @staticmethod
    def from_size(width, length, x_shift=0) -> RobotRenderer:
        return RobotRenderer([
            [-length / 2 + x_shift, -width / 2],
            [+length / 2 + x_shift, -width / 2],
            [+length / 2 + x_shift, +width / 2],
            [-length / 2 + x_shift, +width / 2],
        ])

    def render(self, pixel_size, yaw=0) -> np.ndarray:
        radius = np.linalg.norm(self.outline, axis=1).max()
        width = 2 * int(np.ceil(radius / pixel_size)) + 1

        R = np.array([[np.cos(yaw), -np.sin(yaw)], [np.sin(yaw), np.cos(yaw)]])
        self.rendered_outline = np.array(self.outline).dot(R.T) / pixel_size + width // 2

        renderer = BinaryRenderer((width, width))
        renderer.map.fill(False)
        renderer.polygon(self.rendered_outline)
        return renderer.map



from functools import wraps
from typing import Callable, ParamSpec, TypeVar

from nicegui import ui

_T = TypeVar('_T')
_P = ParamSpec('_P')


class Track:

    def __init__(self) -> None:
        self.stack: list[str] = []

    def __call__(self, f: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(f)
        async def wrap(*args, **kwargs):
            try:
                self.stack.append(f.__name__)
                return await f(*args, **kwargs)
            finally:
                if self.stack and self.stack[-1] == f.__name__:
                    self.stack.pop()
        return wrap

    def ui(self) -> ui.label:
        label = ui.label()
        ui.timer(0.5, lambda: label.set_text(' → '.join(self.stack)))
        return label

    def reset(self) -> None:
        self.stack.clear()


track = Track()



import abc
from typing import Any

from . import registry


class PersistentModule(abc.ABC):
    USE_PERSISTENCE: bool = True

    def __init__(self, *, persistence_key: str | None = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.needs_backup: bool = False
        if self.USE_PERSISTENCE:
            registry.register(self, persistence_key)

    @abc.abstractmethod
    def backup(self) -> dict[str, Any]:
        """Return a JSON-serializable dict with all data that should be backed up."""

    @abc.abstractmethod
    def restore(self, data: dict[str, Any]) -> None:
        """Restore data from a JSON-serializable dict."""

    def request_backup(self) -> None:
        """Mark this module as changed and in need of a backup."""
        self.needs_backup = True



import json
from pathlib import Path
from typing import Callable, Optional

from nicegui import app, events, ui
from starlette.responses import FileResponse

from ..helpers import invoke
from . import registry


def export_button(title: str = 'Export', route: str = '/export', tmp_filepath: Path = Path('/tmp/export.json')) -> ui.button:
    @app.get(route)
    def get_export() -> FileResponse:
        data = {name: module.backup() for name, module in registry.modules.items()}
        tmp_filepath.write_text(json.dumps(data, indent=4))
        return FileResponse(tmp_filepath, filename='export.json')
    return ui.button(title, on_click=lambda: ui.download(route[1:]))


def import_button(title: str = 'Import', after_import: Optional[Callable] = None) -> ui.button:
    async def restore_from_file(e: events.UploadEventArguments) -> None:
        all_data = json.load(e.content)
        assert isinstance(all_data, dict)
        for name, data in all_data.items():
            registry.modules[name].restore(data)
        await registry.backup(force=True)
        dialog.close()
        if after_import is not None:
            await invoke(after_import)
    with ui.dialog() as dialog, ui.card():
        ui.upload(on_upload=restore_from_file)
    return ui.button(title, on_click=dialog.open)



import asyncio
import multiprocessing
from typing import AsyncGenerator

import pytest
from nicegui import core

import rosys
from rosys.analysis import KpiLogger
from rosys.automation import Automator
from rosys.driving import Driver, Odometer
from rosys.geometry import Prism
from rosys.hardware import Robot, RobotSimulation, Wheels, WheelsSimulation
from rosys.pathplanning import PathPlanner
from rosys.testing import helpers, log_configuration

log_configuration.setup()


@pytest.fixture
async def integration() -> AsyncGenerator:
    core.loop = asyncio.get_event_loop()
    rosys.reset_before_test()
    await rosys.startup()
    yield
    await rosys.shutdown()
    rosys.reset_after_test()


@pytest.fixture
async def odometer(wheels: Wheels, integration: None) -> Odometer:
    helpers.odometer = Odometer(wheels)
    return helpers.odometer


@pytest.fixture
async def wheels(integration: None) -> Wheels:
    return WheelsSimulation()


@pytest.fixture
async def robot(wheels: Wheels, integration: None) -> Robot:
    return RobotSimulation([wheels])


@pytest.fixture
async def driver(wheels: Wheels, odometer: Odometer, integration: None) -> Driver:
    helpers.driver = Driver(wheels, odometer)
    return helpers.driver


@pytest.fixture
async def automator(wheels: Wheels, integration: None) -> Automator:
    helpers.automator = Automator(None, on_interrupt=wheels.stop)
    return helpers.automator


@pytest.fixture
def shape() -> Prism:
    return Prism.default_robot_shape()


@pytest.fixture
async def path_planner(shape: Prism, integration: None) -> PathPlanner:
    return PathPlanner(shape)


@pytest.fixture
async def kpi_logger(integration: None) -> KpiLogger:
    return KpiLogger()


@pytest.fixture(scope='session', autouse=True)
def enforce_spawn_process() -> None:
    if multiprocessing.get_start_method() != 'spawn':
        multiprocessing.set_start_method('spawn', force=True)



import asyncio
import logging
from asyncio import Task
from typing import AsyncGenerator, Optional

import httpx

from ...rosys import on_startup
from ..image_processing import remove_exif
from .motec_settings_interface import MotecSettingsInterface
from .vendors import VendorType, mac_to_url, mac_to_vendor


class MjpegDevice:

    def __init__(self, mac: str, ip: str, *,
                 index: Optional[int] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 control_port: int = 8885) -> None:
        self.mac = mac
        self.ip = ip
        self.capture_task: Optional[Task] = None
        self._image_buffer: Optional[bytearray] = None
        self.authentication = None if username is None or password is None else httpx.DigestAuth(username, password)
        self.log = logging.getLogger('rosys.mjpeg_device ' + self.mac)
        url = mac_to_url(mac, ip, index=index)
        if url is None:
            raise ValueError(f'could not determine URL for {mac}')
        self.url = url

        if mac_to_vendor(mac) == VendorType.MOTEC:
            self.settings_interface = MotecSettingsInterface(ip, port=control_port)

        self.start_capture_task()

    def start_capture_task(self) -> None:
        def create_capture_task() -> None:
            loop = asyncio.get_event_loop()
            self.capture_task = loop.create_task(self.run_capture_task())
        on_startup(create_capture_task)

    async def restart_capture(self) -> None:
        self.shutdown()
        self.start_capture_task()

    async def run_capture_task(self) -> None:
        self.log.info('Capturing images from %s', self.url)

        async def stream() -> AsyncGenerator[bytearray, None]:
            async with httpx.AsyncClient() as client:
                assert self.url is not None
                try:
                    async with client.stream('GET', self.url, auth=self.authentication) as response:  # type: ignore
                        if response.status_code != 200:
                            self.log.error('could not connect to %s (credentials: %s): %s %s',
                                           self.url, self.authentication, response.status_code, response.reason_phrase)
                            return

                        buffer_size = 16 * 1024 * 1024
                        buffer = bytearray(buffer_size)
                        buffer_view = memoryview(buffer)
                        buffer_end = 0
                        header = None

                        byte_search_pos = 0

                        try:
                            async for chunk in response.aiter_bytes():
                                chunk_len = len(chunk)

                                if buffer_end + chunk_len > buffer_size:
                                    self.log.warning('Buffer overflow, resetting buffer')
                                    buffer_end = 0
                                    header = None
                                    continue
                                buffer_view[buffer_end:buffer_end + chunk_len] = chunk
                                buffer_end += chunk_len

                                while True:
                                    if header is None:
                                        header_pos = buffer.find(b'\xff\xd8', byte_search_pos, buffer_end)
                                        if header_pos == -1:
                                            break
                                        byte_search_pos = header_pos + 2
                                        header = header_pos
                                    else:
                                        footer_pos = buffer.find(b'\xff\xd9', byte_search_pos, buffer_end)
                                        if footer_pos == -1:
                                            break

                                        image_end = footer_pos + 2
                                        image_data = buffer[header:image_end]
                                        yield image_data

                                        buffer_view[:buffer_end - image_end] = buffer_view[image_end:buffer_end]
                                        header = None
                                        buffer_end -= footer_pos + 2
                                        byte_search_pos = 0
                        except httpx.ReadTimeout:
                            self.log.warning('Connection to %s timed out', self.url)
                except Exception as e:
                    self.log.warning('Connection to %s failed. Was something disconnected?\n%s', self.url, e)
                    raise e

        async for image in stream():
            self._image_buffer = image
        self.log.warning('Capture task stopped')
        self.capture_task = None

    def capture(self) -> Optional[bytes]:
        image = self._image_buffer
        self._image_buffer = None
        return remove_exif(image) if image is not None else None

    def shutdown(self) -> None:
        if self.capture_task is not None:
            self.capture_task.cancel()
            self.capture_task = None



import abc
import logging
from typing import Optional


class Communication(abc.ABC):
    """This abstract module defines an interface for communicating with a microcontroller.

    Besides sending and receiving messages a communication module provides a property whether communication is possible.
    It can also provide a piece of debug UI.
    """

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.communication')

    @classmethod
    @abc.abstractmethod
    def is_possible(cls) -> bool:
        return False

    def connect(self) -> None:
        pass

    def disconnect(self) -> None:
        pass

    @abc.abstractmethod
    async def send(self, msg: str) -> None:
        pass

    @abc.abstractmethod
    async def read(self) -> Optional[str]:
        pass

    def debug_ui(self) -> None:
        pass



from nicegui import ui
from pyloot import PyLoot
from pyloot.server import PyLootServer
from starlette.middleware.wsgi import WSGIMiddleware
from starlette.routing import Mount


class PylootPage():

    def __init__(self) -> None:
        self.pyloot = PyLoot(server=PyLootServer(disable_response_gzip=True))  # gzip is done by starlette
        ui.timer(30, self.pyloot.collect_objects)
        Mount('/pyloot', app=WSGIMiddleware(self.pyloot_wrapper))

    def pyloot_wrapper(self, wsgi_environ, start_response):
        pyloot_environ = wsgi_environ.copy()
        pyloot_environ['wsgi.multiprocess'] = False
        wsgi = self.pyloot.get_wsgi()
        return wsgi(pyloot_environ, start_response)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, Optional

import cv2
import numpy as np
import PIL.Image
import PIL.ImageDraw
from typing_extensions import Self

from .detections import Detections


@dataclass(slots=True, kw_only=True)
class ImageSize:
    width: int
    height: int

    @property
    def tuple(self) -> tuple[int, int]:
        return (self.width, self.height)


@dataclass(slots=True, kw_only=True)
class Image:
    camera_id: str
    size: ImageSize
    time: float  # time of recording
    data: Optional[bytes] = None
    _detections: dict[str, Detections] = field(default_factory=dict)
    is_broken: Optional[bool] = None
    tags: set[str] = field(default_factory=set)

    DEFAULT_PLACEHOLDER_SIZE: ClassVar[tuple[int, int]] = (320, 240)

    @property
    def detections(self) -> Optional[Detections]:
        if not self._detections:
            return None
        if len(self._detections) > 1:
            raise RuntimeError(
                f'Image has multiple detection types ({", ".join(self._detections.keys())}). Use `get_detections(type)` instead.')
        return list(self._detections.values())[0]

    def get_detections(self, detector_id: str) -> Optional[Detections]:
        return self._detections.get(detector_id)

    def set_detections(self, detector_id: str, detections: Detections) -> None:
        self._detections[detector_id] = detections

    @property
    def id(self) -> str:
        return f'{self.camera_id}/{self.time}'

    @classmethod
    def create_placeholder(cls, text: str, time: Optional[float] = None, camera_id: Optional[str] = None, shrink: int = 1) -> Self:
        h, w = cls.DEFAULT_PLACEHOLDER_SIZE
        img = PIL.Image.new('RGB', (h // shrink, w // shrink), color=(73, 109, 137))
        d = PIL.ImageDraw.Draw(img)
        d.text((img.width / 2 - len(text) * 3, img.height / 2 - 5), text, fill=(255, 255, 255))
        _, encoded_image = cv2.imencode('.png', np.array(img)[:, :, ::-1])  # NOTE: cv2 expects BGR
        return cls(
            camera_id=camera_id or 'no_cam_id',
            time=time or 0,
            size=ImageSize(width=img.width, height=img.height),
            data=encoded_image.tobytes(),
        )



from typing import Any

from . import persistence


class Config(persistence.PersistentModule):

    def __init__(self) -> None:
        super().__init__()
        self.ui_update_interval: float = 0.1
        self.simulation_speed: float = 1.0

    def backup(self) -> dict[str, Any]:
        return {'simulation_speed': self.simulation_speed}

    def restore(self, data: dict[str, Any]) -> None:
        self.simulation_speed = data.get('simulation_speed', 1.0)



from __future__ import annotations

import abc
import asyncio
from collections import deque
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Optional

from ... import rosys
from ...event import Event
from ..image import Image
from ..image_route import create_image_route


class Camera(abc.ABC):
    MAX_IMAGES = 256

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: Optional[str] = None,
                 connect_after_init: bool = True,
                 streaming: bool = False,
                 polling_interval: float = 0.1,
                 base_path_overwrite: Optional[str] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.id: str = id
        self.name = name or self.id
        self.connect_after_init = connect_after_init
        self.images: deque[Image] = deque(maxlen=self.MAX_IMAGES)
        self.base_path: str = f'images/{base_path_overwrite or id}'

        self.NEW_IMAGE: Event = Event()

        self.device_connection_lock: asyncio.Condition = asyncio.Condition()

        create_image_route(self)

        if connect_after_init:
            if asyncio.get_event_loop().is_running():
                asyncio.create_task(self.connect())
            else:
                rosys.on_startup(self.connect)

        self.image_loop = rosys.Repeater(self.capture_image, interval=polling_interval)
        if streaming:
            self.image_loop.start()

    def __del__(self) -> None:
        self.image_loop.stop()

    @property
    def streaming(self) -> bool:
        return self.image_loop.running

    @streaming.setter
    def streaming(self, value: bool) -> None:
        if value:
            self.image_loop.start()
        else:
            self.image_loop.stop()

    @property
    def polling_interval(self) -> float:
        return self.image_loop.interval

    @polling_interval.setter
    def polling_interval(self, value: float) -> None:
        self.image_loop.interval = value

    def get_image_url(self, image: Image) -> str:
        return f'{self.base_path}/{image.time}'

    def get_latest_image_url(self) -> str:
        image = self.latest_captured_image
        if image is None:
            return f'{self.base_path}/placeholder'
        return self.get_image_url(image)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'connect_after_init': self.connect_after_init,
            'streaming': self.streaming,
        }

    @property
    def is_connected(self) -> bool:
        """To be interpreted as "ready to capture images"."""
        return False

    @asynccontextmanager
    async def _device_connection(self) -> AsyncGenerator[None, None]:
        await self.device_connection_lock.acquire()
        try:
            yield
        finally:
            self.device_connection_lock.release()

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass

    async def reconnect(self) -> None:
        await self.disconnect()
        await self.connect()

    @property
    def captured_images(self) -> list[Image]:
        return [i for i in self.images if i.data]

    @property
    def latest_captured_image(self) -> Optional[Image]:
        return next((i for i in reversed(self.captured_images) if i.data), None)

    @property
    def latest_detected_image(self) -> Optional[Image]:
        return next((i for i in reversed(self.captured_images) if i.detections), None)

    def get_recent_images(self, *, current_time: Optional[float] = None, timespan: float = 10.0) -> list[Image]:
        """Returns all images that were captured. Latest images are at the end of the list.

        :param current_time: the starting time for the search; defaults to the current time
        :param timespan: the timespan to search back in seconds
        """
        if current_time is None:
            current_time = rosys.time()
        return [i for i in self.captured_images if i.time > current_time - timespan]

    def _add_image(self, image: Image) -> None:
        self.images.append(image)
        self.NEW_IMAGE.emit(image)

    async def capture_image(self) -> None:
        raise NotImplementedError('Implement capture_image() in your camera class!')



from nicegui import ui

from ..rosys import config
from .automator import Automator


class AutomationControls:
    """This UI element contains start/stop/pause/resume buttons for controlling a given automator.

    See [Play-pause-stop](https://rosys.io/examples/play-pause-stop/) for a simple example of the automation controls.
    """

    def __init__(self, automator: Automator) -> None:
        play_button = ui.button(on_click=automator.start) \
            .props('icon=play_arrow unelevated').tooltip('start automation')
        pause_button = ui.button(on_click=lambda: automator.pause(because='pause button was pressed')) \
            .props('icon=pause outline').tooltip('pause automation')
        resume_button = ui.button(on_click=automator.resume) \
            .props('icon=play_arrow outline').tooltip('resume automation')
        stop_button = ui.button(on_click=lambda: automator.stop(because='stop button was pressed')) \
            .props('icon=stop outline').tooltip('stop automation')

        def refresh() -> None:
            play_button.visible = automator.is_stopped
            pause_button.visible = automator.is_running
            resume_button.visible = automator.is_paused
            play_button.enabled = automator.default_automation is not None and automator.enabled
            resume_button.enabled = automator.enabled
            stop_button.enabled = not automator.is_stopped

        ui.timer(config.ui_update_interval, refresh)



from typing import Optional

from nicegui.elements.scene_objects import Group, Object3D

from .area import Area
from .area_manipulation import AreaManipulation, AreaManipulationMode
from .path_planner import PathPlanner

HEIGHT = 0.1
DIAMETER = 0.2


class AreaObject(Group):

    def __init__(self, path_planner: PathPlanner, area_manipulation: Optional[AreaManipulation] = None) -> None:
        super().__init__()

        self.path_planner = path_planner
        self.area_manipulation = area_manipulation

        path_planner.AREAS_CHANGED.register_ui(self.update)
        if area_manipulation:
            area_manipulation.MODE_CHANGED.register_ui(lambda _: self.update())

        self.update()

    def update(self, areas: Optional[list[Area]] = None) -> None:
        area_ids = [area.id for area in areas] if areas else []

        def should_remove(obj: Object3D) -> bool:
            words = (obj.name or '').split('_')
            return (words[0] == 'area') and not (area_ids and words[1] not in area_ids)
        self.scene.delete_objects(should_remove)

        for area in self.path_planner.areas.values():
            if area_ids and area.id not in area_ids:
                continue

            if area.closed:
                outline = [[point.x, point.y] for point in area.outline]
            else:
                outline = [[point.x, point.y] for point in area.outline[:-1] + area.outline[-1:0:-1]]
                last_edge = [[area.outline[0].x, area.outline[0].y], [area.outline[-1].x, area.outline[-1].y]]
                self.scene.extrusion(last_edge, HEIGHT, wireframe=True) \
                    .with_name(f'area_{area.id}_last_edge').material(area.color, 0.2)
            self.scene.extrusion(outline, HEIGHT, wireframe=True).with_name(f'area_{area.id}').material(area.color)

            if self.area_manipulation and self.area_manipulation.mode == AreaManipulationMode.EDIT:
                for p, point in enumerate(area.outline):
                    self.scene.sphere(DIAMETER / 2).move(point.x, point.y, HEIGHT / 2) \
                        .with_name(f'area_{area.id}_corner_{p}').material(area.color).draggable()
                    next_point = area.outline[(p + 1) % len(area.outline)]
                    mid_point = point.interpolate(next_point, 0.5)
                    self.scene.sphere(DIAMETER / 2).move(mid_point.x, mid_point.y, HEIGHT / 2) \
                        .with_name(f'area_{area.id}_mid_{p}').material(area.color, 0.2).draggable()



import asyncio
import logging
import shlex
import subprocess
from asyncio.subprocess import Process
from io import BytesIO
from typing import AsyncGenerator, Optional

from nicegui import background_tasks

from .jovision_rtsp_interface import JovisionInterface
from .vendors import VendorType, mac_to_url, mac_to_vendor


class RtspDevice:

    def __init__(self, mac: str, ip: str, jovision_profile: int) -> None:
        self.log = logging.getLogger('rosys.vision.rtsp_camera.rtsp_device')
        self.log.setLevel(logging.DEBUG)

        self.mac = mac

        self.capture_task: Optional[asyncio.Task] = None
        self.capture_process: Optional[Process] = None
        self._image_buffer: Optional[bytes] = None
        self._authorized: bool = True

        vendor_type = mac_to_vendor(mac)

        self.settings_interface: Optional[JovisionInterface] = None
        if vendor_type == VendorType.JOVISION:
            self.settings_interface = JovisionInterface(ip)
            self.fps = self.settings_interface.get_fps(stream_id=jovision_profile)
        else:
            self.log.warning('[%s] No settings interface for vendor type %s', self.mac, vendor_type)
            self.log.warning('[%s] Using default fps of 10', self.mac)
            self.fps = 10

        url = mac_to_url(mac, ip, jovision_profile)
        if url is None:
            raise ValueError(f'could not determine RTSP URL for {mac}')
        self.url = url
        self.log.info('[%s] Starting VideoStream for %s', self.mac, self.url)
        self.start_gstreamer_task()

    @property
    def authorized(self) -> bool:
        return self._authorized

    def capture(self) -> Optional[bytes]:
        image = self._image_buffer
        self._image_buffer = None
        return image

    def shutdown(self) -> None:
        if self.capture_process is not None:
            self.log.debug('[%s] Terminating gstreamer process', self.mac)
            self.capture_process.terminate()
            self.capture_process = None

    def start_gstreamer_task(self) -> None:
        self.capture_task = background_tasks.create(self.run_gstreamer(self.url), name=f'capture {self.mac}')

    def restart_gstreamer(self) -> None:
        self.shutdown()
        self.start_gstreamer_task()

    async def run_gstreamer(self, url: str) -> None:
        async def stream(url: str) -> AsyncGenerator[bytes, None]:
            if 'subtype=0' in url:
                url = url.replace('subtype=0', 'subtype=1')

            self.log.debug('[%s] Starting gstreamer pipeline for %s', self.mac, url)
            # to try: replace avdec_h264 with nvh264dec ! nvvidconv (!videoconvert)
            command = f'gst-launch-1.0 rtspsrc location="{url}" latency=0 protocols=tcp ! rtph264depay ! avdec_h264 ! videoconvert ! videorate ! "video/x-raw,framerate={self.fps}/1" ! jpegenc ! fdsink'
            process = await asyncio.create_subprocess_exec(*shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            assert process.stdout is not None
            assert process.stderr is not None
            self.capture_process = process

            buffer = BytesIO()
            pos = 0
            header = None

            while process.returncode is None:
                assert process.stdout is not None
                new = await process.stdout.read(4096)
                if not new:
                    break
                buffer.write(new)

                img_range = None
                while True:
                    if header is None:
                        h = buffer.getvalue().find(b'\xff\xd8', pos)
                        if h == -1:
                            pos = buffer.tell() - 1
                            break
                        pos = h + 2
                        header = h
                    else:
                        f = buffer.getvalue().find(b'\xff\xd9', pos)
                        if f == -1:
                            pos = buffer.tell() - 1
                            break
                        img_range = (header, f + 2)
                        pos = f + 2
                        header = None

                if img_range:
                    yield buffer.getvalue()[img_range[0]:img_range[1]]

                    rest = buffer.getvalue()[img_range[1]:]
                    buffer.seek(0)
                    buffer.truncate()
                    buffer.write(rest)

                    pos = 0
                    if header is not None:
                        header -= img_range[1]

            assert process.stderr is not None
            error = await process.stderr.read()
            logging.info('process %s exited with %s and error %s', process.pid, process.returncode, error.decode())
            if 'Unauthorized' in error.decode():
                self._authorized = False

        async for image in stream(url):
            self._image_buffer = image

        self.log.info('[%s] stream ended', self.mac)

        self.capture_task = None



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Optional, Protocol

import numpy as np

from .. import analysis, rosys
from ..geometry import Point, Pose, Spline
from ..helpers import ModificationContext, eliminate_2pi, eliminate_pi, ramp
from .drivable import Drivable
from .odometer import Odometer
from .path_segment import PathSegment


@dataclass(slots=True, kw_only=True)
class DriveParameters(ModificationContext):
    linear_speed_limit: float = 0.5
    angular_speed_limit: float = 0.5
    minimum_turning_radius: float = 0.0
    can_drive_backwards: bool = True
    max_detection_age_ramp: Optional[tuple[float, float]] = None
    hook_offset: float = 0.5
    carrot_offset: float = 0.6
    carrot_distance: float = 0.1
    hook_bending_factor: float = 0


@dataclass(slots=True, kw_only=True)
class DriveState:
    carrot_pose: Pose
    curvature: float
    backward: bool
    turn_angle: float


class DrivingAbortedException(Exception):
    pass


class PoseProvider(Protocol):

    @property
    def pose(self) -> Pose:
        ...


class Driver:
    """The driver module allows following a given path.

    It requires a wheels module (or any drivable hardware representation) to execute individual drive commands.
    It also requires an odometer to get a current prediction of the robot's pose.
    Its `parameters` allow controlling the specific drive behavior.
    """

    def __init__(self, wheels: Drivable, odometer: Odometer | PoseProvider) -> None:
        self.wheels = wheels
        self.odometer = odometer
        self.parameters = DriveParameters()
        self.state: Optional[DriveState] = None
        self._abort = False

    @property
    def prediction(self) -> Pose:
        """The current prediction of the robot's pose based on the odometer."""
        return self.odometer.prediction if isinstance(self.odometer, Odometer) else self.odometer.pose

    def abort(self) -> None:
        """Abort the current drive routine."""
        self._abort = True

    @analysis.track
    async def drive_square(self) -> None:
        start_pose = deepcopy(self.prediction)
        for x, y in [(1, 0), (1, 1), (0, 1), (0, 0)]:
            await self.drive_to(start_pose.transform(Point(x=x, y=y)))

    @analysis.track
    async def drive_arc(self) -> None:
        while self.prediction.x < 2:
            if self._abort:
                self._abort = False
                raise DrivingAbortedException()
            await self.wheels.drive(1, np.deg2rad(25))
            await rosys.sleep(0.1)
        await self.wheels.stop()

    @analysis.track
    async def drive_path(self, path: list[PathSegment]) -> None:
        for segment in path:
            await self.drive_spline(segment.spline, throttle_at_end=segment == path[-1], flip_hook=segment.backward)

    @analysis.track
    async def drive_to(self, target: Point, backward: bool = False) -> None:
        if self.parameters.minimum_turning_radius:
            await self.drive_circle(target, backward)

        robot_position = self.prediction.point
        approach_spline = Spline(
            start=robot_position,
            control1=robot_position.interpolate(target, 1/3),
            control2=robot_position.interpolate(target, 2/3),
            end=target,
        )
        await self.drive_spline(approach_spline, flip_hook=backward)

    @analysis.track
    async def drive_circle(self, target: Point, backward: bool = False) -> None:
        while True:
            if self._abort:
                self._abort = False
                raise DrivingAbortedException()
            target_yaw = self.prediction.direction(target)
            if backward:
                target_yaw += np.pi
            angle = eliminate_2pi(target_yaw - self.prediction.yaw)
            if abs(angle) < np.deg2rad(5):
                break
            linear = 0.5 if not backward else -0.5
            sign = 1 if angle > 0 else -1
            if backward:
                sign *= -1
            angular = linear / self.parameters.minimum_turning_radius * sign
            await self.wheels.drive(*self._throttle(linear, angular))
            await rosys.sleep(0.1)

    @analysis.track
    async def drive_spline(self, spline: Spline, *, flip_hook: bool = False, throttle_at_end: bool = True) -> None:
        if spline.start.distance(spline.end) < 0.01:
            return  # NOTE: skip tiny splines

        hook_offset = Point(x=self.parameters.hook_offset, y=0) * (-1 if flip_hook else 1)
        carrot_offset = Point(x=self.parameters.carrot_offset, y=0)
        carrot = Carrot(spline=spline, offset=carrot_offset)

        while True:
            if self._abort:
                self._abort = False
                raise DrivingAbortedException()
            velocity = self.odometer.current_velocity if isinstance(self.odometer, Odometer) else None
            dYaw = self.parameters.hook_bending_factor * velocity.angular if velocity else 0
            hook = self.prediction.transform_pose(Pose(yaw=dYaw)).transform(hook_offset)
            if self.parameters.can_drive_backwards:
                can_move = carrot.move(hook, distance=self.parameters.carrot_distance)
            else:
                can_move = carrot.move_by_foot(self.prediction)
            if not can_move:
                break

            turn_angle = eliminate_pi(hook.direction(carrot.offset_point) - self.prediction.yaw)
            curvature = np.tan(turn_angle) / hook_offset.x
            if curvature != 0 and abs(1 / curvature) < self.parameters.minimum_turning_radius:
                curvature = (-1 if curvature < 0 else 1) / self.parameters.minimum_turning_radius

            drive_backward = hook.projected_distance(carrot.offset_point, self.prediction.yaw) < 0
            if drive_backward and not self.parameters.can_drive_backwards:
                drive_backward = False
                curvature = (-1 if curvature > 0 else 1) / max(self.parameters.minimum_turning_radius, 0.001)
            linear: float = -1 if drive_backward else 1
            t = spline.closest_point(hook.x, hook.y)
            if t >= 1.0 and throttle_at_end:
                target_distance = self.prediction.projected_distance(spline.pose(1.0))
                linear *= ramp(target_distance, self.parameters.hook_offset, 0.0, 1.0, 0.01, clip=True)
            angular = linear * curvature

            self.state = DriveState(
                carrot_pose=carrot.pose,
                curvature=curvature,
                backward=drive_backward,
                turn_angle=turn_angle,
            )

            await self.wheels.drive(*self._throttle(linear, angular))
            await rosys.sleep(0.1)

        self.state = None
        await self.wheels.stop()

    def _throttle(self, linear: float, angular: float) -> tuple[float, float]:
        factor = self.throttle_factor()
        linear *= factor
        angular *= factor

        linear_limit = self.parameters.linear_speed_limit
        angular_limit = self.parameters.angular_speed_limit

        if abs(linear) > linear_limit:
            factor = linear_limit / abs(linear)
            angular *= factor
            linear *= factor
        if abs(angular) > angular_limit:
            factor = angular_limit / abs(angular)
            linear *= factor
            angular *= factor

        return linear, angular

    def throttle_factor(self) -> float:
        if self.parameters.max_detection_age_ramp is None:
            return 1
        if not isinstance(self.odometer, Odometer):
            raise ValueError('max_detection_age_ramp requires an odometer')
        if self.odometer.detection is None:
            return 0
        age_ramp = self.parameters.max_detection_age_ramp
        age = rosys.time() - self.odometer.detection.time
        return ramp(age, age_ramp[0], age_ramp[1], 1.0, 0.0, clip=True)


@dataclass(slots=True, kw_only=True)
class Carrot:
    spline: Spline
    offset: Point = field(default_factory=lambda: Point(x=0, y=0))
    t: float = 0
    _estimated_spline_length: float = field(init=False)

    def __post_init__(self) -> None:
        self._estimated_spline_length = self.spline.estimated_length()  # NOTE: compute once to avoid repeated computation

    @property
    def pose(self) -> Pose:
        if self.t < 1.0:
            return self.spline.pose(self.t)
        return Pose(
            x=self.spline.x(1.0) + (self.t - 1.0) * self.spline.gx(1.0),
            y=self.spline.y(1.0) + (self.t - 1.0) * self.spline.gy(1.0),
            yaw=self.spline.yaw(1.0),
        )

    @property
    def offset_point(self) -> Point:
        return self.pose.transform(self.offset)

    def move(self, hook: Point, distance: float) -> bool:
        dt = 0.1 * distance / self._estimated_spline_length
        while hook.distance(self.offset_point) < distance:
            self.t += dt
            if self.t >= 1.0:
                return False
        return True

    def move_by_foot(self, pose: Pose) -> bool:
        self.t = self.spline.closest_point(pose.x, pose.y, t_min=self.t, t_max=1.0)
        return self.t < 1.0



from __future__ import annotations

from dataclasses import dataclass

import cv2
import numpy as np
from pyquaternion import Quaternion


@dataclass(slots=True, kw_only=True)
class Rotation:
    R: list[list[float]]

    @staticmethod
    def zero() -> Rotation:
        return Rotation(R=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    @staticmethod
    def from_euler(roll: float, pitch: float, yaw: float) -> Rotation:
        Rx = np.array([[1, 0, 0], [0, np.cos(roll), -np.sin(roll)], [0, np.sin(roll), np.cos(roll)]])
        Ry = np.array([[np.cos(pitch), 0, np.sin(pitch)], [0, 1, 0], [-np.sin(pitch), 0, np.cos(pitch)]])
        Rz = np.array([[np.cos(yaw), -np.sin(yaw), 0], [np.sin(yaw), np.cos(yaw), 0], [0, 0, 1]])
        return Rotation(R=(Rz @ Ry @ Rx).tolist())

    @staticmethod
    def from_quaternion(w: float, x: float, y: float, z: float) -> Rotation:
        return Rotation(R=Quaternion(w, x, y, z).rotation_matrix.tolist())

    @staticmethod
    def from_rvec(rvec) -> Rotation:
        return Rotation(R=cv2.Rodrigues(rvec)[0].tolist())

    def __mul__(self, other) -> Rotation:
        return Rotation(R=np.dot(self.R, other.R).tolist())

    @property
    def T(self) -> Rotation:
        return Rotation(R=np.array(self.R).T.tolist())

    @property
    def roll(self) -> float:
        return np.arctan2(self.R[2][1], self.R[2][2])

    @property
    def pitch(self) -> float:
        return np.arctan2(-self.R[2][0], np.sqrt(self.R[2][1]**2 + self.R[2][2]**2))

    @property
    def yaw(self) -> float:
        return np.arctan2(self.R[1][0], self.R[0][0])

    @property
    def euler(self) -> tuple[float, float, float]:
        return self.roll, self.pitch, self.yaw

    @property
    def quaternion(self) -> tuple[float, float, float, float]:
        return Quaternion(matrix=np.array(self.R)).elements

    @property
    def total_angle(self) -> float:
        # https://en.wikipedia.org/wiki/Rotation_matrix#Determining_the_angle
        trace = np.trace(self.R)
        return 0 if np.isclose(trace, 3) else np.arccos((trace - 1) / 2)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        roll, pitch, yaw = np.round(np.rad2deg(self.euler), 2)
        return f'{roll:6.1f} {pitch:6.1f} {yaw:6.1f}'



from typing import Optional

import numpy as np
from typing_extensions import Self

from ...geometry import Rotation
from ..calibration import Calibration, Extrinsics, Intrinsics
from .camera import Camera


class CalibratableCamera(Camera):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.calibration: Optional[Calibration] = None

    @property
    def is_calibrated(self) -> bool:
        return self.calibration is not None

    @classmethod
    def create_calibrated(cls, *,
                          width: int = 800, height: int = 600,
                          x: float = 0.0, y: float = 0.0, z: float = 1.0,
                          roll: float = np.pi, pitch: float = 0.0, yaw: float = 0.0,
                          **kwargs) -> Self:
        camera = cls(**kwargs)
        camera.set_perfect_calibration(width=width, height=height, x=x, y=y, z=z, roll=roll, pitch=pitch, yaw=yaw)
        return camera

    def set_perfect_calibration(self, *,
                                width=800, height=600,
                                x: float = 0.0, y: float = 0.0, z: float = 1.0,
                                roll: float = np.pi, pitch: float = 0.0, yaw: float = 0.0) -> None:
        self.calibration = Calibration(
            intrinsics=Intrinsics.create_default(width, height),
            extrinsics=Extrinsics(rotation=Rotation.from_euler(roll, pitch, yaw), translation=[x, y, z]),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import overload

import numpy as np

from .line import Line
from .point import Point
from .point3d import Point3d


@dataclass(slots=True, kw_only=True)
class PoseStep:
    linear: float
    angular: float
    time: float


@dataclass(slots=True, kw_only=True)
class Pose:
    x: float = 0
    y: float = 0
    yaw: float = 0
    time: float = 0

    @property
    def point(self) -> Point:
        return Point(x=self.x, y=self.y)

    def point_3d(self) -> Point3d:
        return Point3d(x=self.x, y=self.y, z=0)

    @property
    def matrix(self) -> np.ndarray:
        return np.array([
            [np.cos(self.yaw), -np.sin(self.yaw), self.x],
            [np.sin(self.yaw), np.cos(self.yaw), self.y],
            [0, 0, 1],
        ])

    @property
    def inv_matrix(self) -> np.ndarray:
        return np.array([
            [np.cos(self.yaw), np.sin(self.yaw), -np.cos(self.yaw) * self.x - np.sin(self.yaw) * self.y],
            [-np.sin(self.yaw), np.cos(self.yaw), np.sin(self.yaw) * self.x - np.cos(self.yaw) * self.y],
            [0, 0, 1],
        ])

    @property
    def yaw_deg(self) -> float:
        return np.rad2deg(self.yaw)

    @staticmethod
    def from_matrix(M, time: float = 0) -> Pose:
        return Pose(x=M[0, -1], y=M[1, -1], yaw=np.arctan2(M[1, 0], M[0, 0]), time=time)

    def __str__(self) -> str:
        return f'{self.x:.3f}, {self.y:.3f}, {self.yaw_deg:.1f} deg'

    @overload
    def distance(self, other: Point) -> float: ...

    @overload
    def distance(self, other: Pose) -> float: ...

    def distance(self, other: Point | Pose) -> float:
        return float(np.sqrt((other.x - self.x)**2 + (other.y - self.y)**2))

    def projected_distance(self, other: Pose) -> float:
        return self.point.projected_distance(other.point, other.yaw)

    @overload
    def direction(self, other: Point) -> float: ...

    @overload
    def direction(self, other: Pose) -> float: ...

    def direction(self, other: Point | Pose) -> float:
        return float(np.arctan2(other.y - self.y, other.x - self.x))

    def __iadd__(self, step: PoseStep) -> Pose:
        self.x += step.linear * np.cos(self.yaw)
        self.y += step.linear * np.sin(self.yaw)
        self.yaw += step.angular
        self.time = step.time
        return self

    def __add__(self, step: PoseStep) -> Pose:
        return Pose(
            x=self.x + step.linear * np.cos(self.yaw),
            y=self.y + step.linear * np.sin(self.yaw),
            yaw=self.yaw + step.angular,
            time=step.time,
        )

    def transform(self, point: Point) -> Point:
        return Point(
            x=self.x + point.x * np.cos(self.yaw) - point.y * np.sin(self.yaw),
            y=self.y + point.x * np.sin(self.yaw) + point.y * np.cos(self.yaw),
        )

    def transform_array(self, points: np.ndarray) -> np.ndarray:
        return np.stack((
            self.x + points[:, 0] * np.cos(self.yaw) - points[:, 1] * np.sin(self.yaw),
            self.y + points[:, 0] * np.sin(self.yaw) + points[:, 1] * np.cos(self.yaw),
        ), axis=1)

    def transform_pose(self, pose: Pose) -> Pose:
        return Pose(
            x=self.x + pose.x * np.cos(self.yaw) - pose.y * np.sin(self.yaw),
            y=self.y + pose.x * np.sin(self.yaw) + pose.y * np.cos(self.yaw),
            yaw=self.yaw + pose.yaw,
            time=pose.time,
        )

    def transform_line(self, line: Line) -> Line:
        a, b, c = self.inv_matrix.T @ line.tuple
        return Line(a=a, b=b, c=c)

    def relative_pose(self, other: Pose) -> Pose:
        return Pose.from_matrix(self.inv_matrix @ other.matrix)

    def relative_point(self, other: Point) -> Point:
        dx = other.x - self.x
        dy = other.y - self.y
        return Point(x=dx * np.cos(-self.yaw) - dy * np.sin(-self.yaw),
                     y=dx * np.sin(-self.yaw) + dy * np.cos(-self.yaw))

    def rotate(self, angle: float) -> Pose:
        return Pose(x=self.x, y=self.y, yaw=self.yaw+angle, time=self.time)

    def interpolate(self, other: Pose, t: float) -> Pose:
        return Pose(
            x=(1 - t) * self.x + t * other.x,
            y=(1 - t) * self.y + t * other.y,
            yaw=(1 - t) * self.yaw + t * other.yaw,
            time=(1 - t) * self.time + t * other.time,
        )



from __future__ import annotations

from datetime import date, datetime
from itertools import groupby
from typing import Any

from rosys import persistence

from .. import rosys
from .kpi_buckets import Day, Month


def date_to_str(date_: date) -> str:
    return date_.isoformat()


def str_to_date(date_: str) -> datetime:
    return datetime.fromisoformat(date_)


class KpiLogger(persistence.PersistentModule):

    def __init__(self) -> None:
        super().__init__()

        self.states: dict[str, bool] = {}
        self.days: list[Day] = []
        self.months: list[Month] = []

    def backup(self) -> dict:
        self.pack()
        return {
            'days': persistence.to_dict(self.days),
            'months': persistence.to_dict(self.months),
        }

    def restore(self, data: dict[str, Any]) -> None:
        persistence.replace_list(self.days, Day, data.get('days', []))
        persistence.replace_list(self.months, Month, data.get('months', []))

    def pack(self) -> None:
        """pack old days into months (as long as there are at least 3 unpacked months)"""
        while True:
            month_groups = [list(days) for _, days in groupby(self.days, key=lambda d: d.date[:-3])]
            if len(month_groups) <= 4:
                break
            month = Month.from_buckets(month_groups[0])
            self.months.append(month)
            for day in month_groups[0]:
                self.days.remove(day)
            self.request_backup()

    def today(self) -> Day:
        date_ = date_to_str(datetime.utcfromtimestamp(rosys.time()).date())
        if not self.days or self.days[-1].date != date_:
            self.days.append(Day(date=date_))
        return self.days[-1]

    def increment(self, key: str) -> None:
        day = self.today()
        day.incidents[key] = day.incidents.get(key, 0) + 1
        self.request_backup()

    def increment_on_rising_edge(self, key: str, new_state: bool) -> None:
        if new_state and not self.states.get(key, False):
            self.increment(key)
        self.states[key] = new_state



from contextlib import nullcontext

from nicegui.elements.scene_objects import Extrusion, Group

from .obstacle import Obstacle
from .path_planner import PathPlanner


class ObstacleObject(Group):

    def __init__(self, path_planner: PathPlanner):
        super().__init__()

        self.update(path_planner.obstacles)
        path_planner.OBSTACLES_CHANGED.register_ui(self.update)

    def update(self, obstacles: dict[str, Obstacle]) -> None:
        for obj in list(self.scene.objects.values()):
            if (obj.name or '').startswith('obstacle_'):
                obj.delete()
        with self.scene or nullcontext():
            for obstacle in obstacles.values():
                outline = [[point.x, point.y] for point in obstacle.outline]
                Extrusion(outline, 0.1).with_name(f'obstacle_{obstacle.id}')



import logging
from enum import Enum

import numpy as np

from rosys.event import Event
from rosys.helpers import ramp

from .. import rosys
from .drivable import Drivable


class State(Enum):
    IDLE = 1
    STEERING = 2
    STOPPING = 3


class Steerer:
    """The steerer module translates x-y information (e.g. from a joystick) to linear/angular velocities sent to the robot.

    The wheels module can be any drivable hardware representation.
    Changing the steering state emits events that can be used to react to manual user interaction.

    The conversion from x-y joystick coordinates to linear and angular velocities is implemented as follows:
    The coordinates are translated into an angle (forward: 0 degrees, backward: 180 degrees).
    If the angle is below 110 degrees, y is used as linear velocity and x is used as angular velocity:
    Pulling the joystick to the front-right corner leads to a clockwise rotation while driving forwards.
    Above 110 degrees the angular velocity is flipped:
    Pulling the joystick to the rear-right corner leads to a counter-clockwise rotation while driving backwards.
    From 100 degrees to 110 degrees both velocities are throttled with a linear ramp from factor 1.0 down to 0.0.
    From 110 degrees to 120 degrees the throttle factor is linearly ramped from 0.0 back to 1.0.
    """

    def __init__(self, wheels: Drivable, speed_scaling: float = 1.0) -> None:
        self.STEERING_STARTED = Event()
        """steering has started"""

        self.STEERING_STOPPED = Event()
        """steering has stopped"""

        self.log = logging.getLogger('rosys.steerer')

        self.wheels = wheels
        self.speed_scaling = speed_scaling
        self.state = State.IDLE
        self.linear_speed = 0.0
        self.angular_speed = 0.0

        rosys.on_repeat(self.step, 0.05)

    def start(self) -> None:
        self.log.info('start steering')
        self.state = State.STEERING
        self.STEERING_STARTED.emit()

    def update(self, x: float, y: float) -> None:
        if self.state == State.STEERING:
            degrees = np.abs(np.rad2deg(np.arctan2(x, y)))
            throttle = ramp(abs(degrees - 110), 0, 10, 0, 1, clip=True)
            flip = degrees > 110
            self.linear_speed = y * self.speed_scaling * throttle
            self.angular_speed = -x * self.speed_scaling * throttle * (-1 if flip else 1)

    def stop(self) -> None:
        if self.state != State.STEERING:
            return
        self.log.info('stop steering')
        self.state = State.STOPPING
        self.STEERING_STOPPED.emit()

    async def step(self) -> None:
        if self.state == State.STEERING:
            await self.wheels.drive(self.linear_speed, self.angular_speed)
        elif self.state == State.STOPPING:
            await self.wheels.stop()
            self.state = State.IDLE

    def set_speed_scaling(self, speed_scaling: float) -> None:
        self.speed_scaling = speed_scaling

    def __str__(self) -> str:
        return f'{type(self).__name__} ({self.state})'



import numpy as np
from nicegui import ui
from nicegui.elements.scene_object3d import Object3D
from nicegui.elements.scene_objects import Cylinder, Group, Text, Texture

from .. import run
from .calibratable_camera_provider import CalibratableCameraProvider
from .calibration import Calibration
from .camera import CalibratableCamera
from .camera_projector import CameraProjector


class CameraObjects(Group):
    """This module provides a UI element for displaying cameras in a 3D scene.

    It requires a camera provider as a source of cameras as well as a camera projector to show the current images projected on the ground plane.
    The `px_per_m` argument can be used to scale the camera frustums.
    With `debug=True` camera IDs are shown (default: `False`).
    """

    def __init__(self,
                 camera_provider: CalibratableCameraProvider,
                 camera_projector: CameraProjector,
                 *,
                 px_per_m: float = 10000,
                 debug: bool = False,
                 ) -> None:
        super().__init__()

        self.camera_provider = camera_provider
        self.camera_projector = camera_projector
        self.px_per_m = px_per_m
        self.debug = debug
        self.textures: dict[str, Texture] = {}
        self.image_shrink_factor = 2

        ui.timer(1.0, self.update)

    @property
    def calibrated_cameras(self) -> dict[str, CalibratableCamera]:
        return {id: camera for id, camera in self.camera_provider.cameras.items() if camera.calibration}

    def find_objects(self, type_: str) -> dict[str, Object3D]:
        return {
            (obj.name or '').split('_', 1)[1]: obj
            for obj in self.scene.objects.values()
            if (obj.name or '').split('_', 1)[0] == type_
        }

    async def update(self) -> None:
        await self.update_cameras()
        await self.update_images()

    async def update_cameras(self) -> None:
        camera_groups = self.find_objects('camera')
        for uid, camera_group in camera_groups.items():
            if uid not in self.calibrated_cameras:
                camera_group.delete()
        for uid, camera in self.calibrated_cameras.items():
            assert camera.calibration is not None
            if uid not in camera_groups:
                with self:
                    with Group().with_name(f'camera_{uid}') as camera_groups[uid]:
                        with Group() as pyramid:
                            Cylinder(0, np.sqrt(0.5), 1, 4) \
                                .rotate(-np.pi / 2, 0, np.pi / 4) \
                                .move(z=0.5) \
                                .material('#0088ff')
                            if self.debug:
                                Text(uid)
                        pyramid.scale(
                            camera.calibration.intrinsics.size.width / self.px_per_m,
                            camera.calibration.intrinsics.size.height / self.px_per_m,
                            camera.calibration.intrinsics.matrix[0][0] / self.px_per_m,
                        )
            camera_groups[uid].move(*camera.calibration.extrinsics.translation)
            camera_groups[uid].rotate_R(await run.cpu_bound(self.get_rotation, camera.calibration))

    @staticmethod
    def get_rotation(calibration: Calibration) -> list[list[float]]:
        return calibration.rotation.R  # computing euler rotation is cpu expensive

    async def update_images(self) -> None:
        newest_images = [camera.latest_captured_image
                         for camera in self.calibrated_cameras.values()
                         if camera.latest_captured_image is not None]

        current_uids = [image.camera_id for image in newest_images]
        for uid, texture in list(self.textures.items()):
            if uid not in current_uids:
                texture.delete()
                del self.textures[uid]

        z = 0.0
        for image in sorted(newest_images, key=lambda i: i.time):
            camera = self.calibrated_cameras.get(image.camera_id)
            if camera is None:
                continue
            projection = self.camera_projector.projections.get(camera.id)
            if projection is None:
                continue
            coordinates = [[point and [point[0], point[1], 0] for point in row] for row in projection.coordinates]

            url = f'{camera.get_image_url(image)}?shrink={self.image_shrink_factor}'
            if image.camera_id not in self.textures:
                with self:
                    self.textures[image.camera_id] = Texture(url, coordinates).with_name(f'image_{image.id}')
            texture = self.textures[image.camera_id]

            z += 0.001
            texture.move(z=z)
            if texture.args[0] != url:
                texture.set_url(url)
            if not await run.cpu_bound(CameraProjector.allclose, texture.args[1], coordinates):
                texture.set_coordinates(coordinates)



from nicegui import background_tasks

from . import analysis, automation, driving, event, geometry, hardware, pathplanning, persistence, run, system, vision
from .config import Config
from .rosys import (NEW_NOTIFICATION, Notification, config, is_test, notify, on_repeat, on_shutdown, on_startup,
                    reset_after_test, reset_before_test, set_time, shutdown, sleep, startup, time, translator, uptime)
from .simulation_ui import simulation_ui
from .version import __version__

__all__ = [
    'background_tasks',
    'analysis',
    'automation',
    'driving',
    'event',
    'geometry',
    'hardware',
    'pathplanning',
    'persistence',
    'run',
    'system',
    'vision',
    'Config',
    'NEW_NOTIFICATION',
    'Notification',
    'config',
    'is_test',
    'notify',
    'on_repeat',
    'on_shutdown',
    'on_startup',
    'reset_after_test',
    'reset_before_test',
    'set_time',
    'shutdown',
    'sleep',
    'startup',
    'time',
    'translator',
    'uptime',
    'simulation_ui',
    '__version__',
]



from .asyncio_warnings import AsyncioWarnings
from .kpi_buckets import Day, Month, TimeBucket, Week
from .kpi_chart import KpiChart
from .kpi_logger import KpiLogger, date_to_str, str_to_date
from .kpi_page_ import kpi_page
from .legacy.asyncio_monitor import AsyncioMonitor
from .legacy.network_monitor import NetworkMonitor, NetworkStats
from .legacy.objgraph_page import objgraph_page
from .logging_page import LoggingPage as logging_page
from .memory import MemoryMiddleware
from .profile_button_ import ProfileButton as profile_button
from .timelapse_recorder import TimelapseRecorder
from .track import track
from .videos_page_ import VideosPage as videos_page

__all__ = [
    'AsyncioWarnings',
    'Day',
    'Month',
    'TimeBucket',
    'Week',
    'KpiChart',
    'KpiLogger',
    'date_to_str',
    'str_to_date',
    'kpi_page',
    'AsyncioMonitor',
    'NetworkMonitor',
    'NetworkStats',
    'objgraph_page',
    'logging_page',
    'MemoryMiddleware',
    'profile_button',
    'TimelapseRecorder',
    'track',
    'videos_page',
]



from dataclasses import dataclass, field
from typing import Optional
from uuid import uuid4

import numpy as np

from .. import rosys
from ..geometry import Point3d
from .calibratable_camera_provider import CalibratableCameraProvider
from .detections import BoxDetection, Detections, PointDetection
from .detector import Autoupload, Detector
from .image import Image


@dataclass(slots=True, kw_only=True)
class SimulatedObject:
    category_name: str
    position: Point3d
    size: Optional[tuple[float, float, float]] = None
    uuid: str = field(init=False)
    confidence: float = 1.0

    def __post_init__(self) -> None:
        self.uuid = str(uuid4())


class DetectorSimulation(Detector):
    """This detector simulates object detection.

    It requires a camera provider in order to check visibility using the cameras' calibrations.
    Individual camera IDs can be added to a set of `blocked_cameras` to simulate occlusions during pytests.
    A list of `simulated_objects` can be filled to define what can be detected.
    An optional `noise` parameter controls the spatial accuracy in pixels.
    """

    def __init__(self, camera_provider: CalibratableCameraProvider, *, noise: float = 1.0, name: Optional[str] = None) -> None:
        super().__init__(name=name)

        self.camera_provider = camera_provider
        self.noise = noise

        self.blocked_cameras: set[str] = set()
        self.simulated_objects: list[SimulatedObject] = []

        rosys.on_repeat(self.step, 0.1)

    def step(self) -> None:
        self.uploads.queue.clear()
        self.uploads.priority_queue.clear()

    async def detect(self,
                     image: Image,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] = [],
                     ) -> Detections | None:
        is_blocked = image.camera_id in self.blocked_cameras
        await rosys.sleep(0.4)
        image.set_detections(self.name, Detections())
        if not is_blocked:
            self.update_simulated_objects(image)
            self.detect_from_simulated_objects(image)
        self.NEW_DETECTIONS.emit(image)
        return image.get_detections(self.name)

    async def upload(self, image: Image, *, tags: list[str] = []) -> None:
        self.log.info('Uploading %s', image.id)

    def update_simulated_objects(self, image: Image) -> None:
        pass

    def detect_from_simulated_objects(self, image: Image) -> None:
        if image.camera_id not in self.camera_provider.cameras:
            return
        camera = self.camera_provider.cameras[image.camera_id]
        assert camera.calibration is not None
        detections = image.get_detections(self.name)
        assert detections is not None
        for obj in self.simulated_objects:
            viewing_direction = np.array(camera.calibration.extrinsics.rotation.R)[:, 2]
            object_direction = np.array(obj.position.tuple) - camera.calibration.extrinsics.translation
            if np.dot(viewing_direction, object_direction) < 0:
                continue
            image_point = camera.calibration.project_to_image(obj.position)
            if image_point is None:
                continue
            if not (0 <= image_point.x < image.size.width and 0 <= image_point.y < image.size.height):
                continue

            if obj.size is None:
                detections.points.append(PointDetection(
                    category_name=obj.category_name,
                    model_name='simulation',
                    confidence=obj.confidence,
                    x=image_point.x + self.noise * np.random.randn(),
                    y=image_point.y + self.noise * np.random.randn(),
                    uuid=obj.uuid,
                ))
            else:
                world_points = np.array([
                    [obj.position.x + dx, obj.position.y + dy, obj.position.z + dz]
                    for dx in [-obj.size[0] / 2, obj.size[0] / 2]
                    for dy in [-obj.size[1] / 2, obj.size[1] / 2]
                    for dz in [-obj.size[2] / 2, obj.size[2] / 2]
                ])
                image_points = camera.calibration.project_to_image(world_points)
                if np.any(np.isnan(image_points)):
                    continue
                detections.boxes.append(BoxDetection(
                    category_name=obj.category_name,
                    model_name='simulation',
                    confidence=obj.confidence,
                    x=image_points[:, 0].min() + self.noise * np.random.randn(),
                    y=image_points[:, 1].min() + self.noise * np.random.randn(),
                    width=image_points[:, 0].max() - image_points[:, 0].min() + self.noise * np.random.randn(),
                    height=image_points[:, 1].max() - image_points[:, 1].min() + self.noise * np.random.randn(),
                    uuid=obj.uuid,
                ))



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from ..helpers import is_test
from ..run import awaitable

if TYPE_CHECKING:
    from .persistent_module import PersistentModule

log = logging.getLogger('rosys.persistence')
backup_path = Path('~/.rosys').expanduser()
modules: dict[str, PersistentModule] = {}


def register(module: PersistentModule, key: str | None = None) -> None:
    if not is_test():
        modules[key or module.__module__] = module


class Encoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, np.floating):
            return float(o)
        return json.JSONEncoder.default(self, o)


@awaitable
def backup(force: bool = False) -> None:
    for name, module in modules.items():
        if not module.needs_backup and not force:
            continue
        if not backup_path.exists():
            backup_path.mkdir(parents=True)
        filepath = backup_path / f'{name}.json'
        try:
            filepath.write_text(json.dumps(module.backup(), indent=4, cls=Encoder))
        except Exception:
            log.exception('failed to backup %s: %s', module, str(module.backup()))
        module.needs_backup = False


def restore() -> None:
    for name, module in modules.items():
        filepath = backup_path / f'{name}.json'
        if not filepath.exists():
            log.warning('Backup file "%s" not found.', filepath)
            continue
        try:
            module.restore(json.loads(filepath.read_text()))
        except Exception:
            log.exception('failed to restore %s', module)



import functools
import inspect
import logging
import logging.config
import os
import sys
import time
from contextlib import contextmanager
from typing import Any, Awaitable, Callable, Generator, Optional

import numpy as np
from nicegui import app

from .lazy_worker import LazyWorker

__all__ = [
    'LazyWorker',
    'invoke',
    'measure',
    'angle',
    'eliminate_pi',
    'eliminate_2pi',
    'ramp',
    'remove_indentation',
    'ModificationContext',
    'from_dict',
    'is_stopping',
    'is_test',
    'PackagePathFilter',
]


async def invoke(handler: Callable, *args: Any) -> Any:
    result = handler(*args)
    if isinstance(result, Awaitable):
        result = await result
    return result


def measure(*, reset: bool = False, ms: bool = False) -> None:
    if 't' in globals() and not reset:
        dt = time.perf_counter() - globals()['t']
        line = inspect.stack()[1][0].f_lineno
        output = f'{dt * 1000:7.3f} ms' if ms else f'{dt:7.3f} s'
        print(f'{inspect.stack()[1].filename}:{line}', output, flush=True)
    if reset:
        print('------------', flush=True)
    globals()['t'] = time.perf_counter()


def angle(yaw0: float, yaw1: float) -> float:
    return eliminate_2pi(yaw1 - yaw0)


def eliminate_pi(angle_: float) -> float:
    return (angle_ + np.pi / 2) % np.pi - np.pi / 2


def eliminate_2pi(angle_: float) -> float:
    return (angle_ + np.pi) % (2 * np.pi) - np.pi


def ramp(x: float, in1: float, in2: float, out1: float, out2: float, clip: bool = False) -> float:
    """Map a value x from one range (in1, in2) to another (out1, out2)."""
    if clip:
        if x < in1 < in2 or in2 < in1 < x:
            return out1
        if x < in2 < in1 or in1 < in2 < x:
            return out2
    return (x - in1) * (out2 - out1) / (in2 - in1) + out1


def remove_indentation(text: str) -> str:
    """Remove indentation from a multi-line string based on the indentation of the first line."""
    lines = text.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    if not lines:
        return ''
    indentation = len(lines[0]) - len(lines[0].lstrip())
    return '\n'.join(line[indentation:] for line in lines)


class ModificationContext:

    @contextmanager
    def set(self, **kwargs) -> Generator[None, None, None]:
        backup = {key: getattr(self, key) for key in kwargs}
        for key, value in kwargs.items():
            setattr(self, key, value)
        try:
            yield
        finally:
            for key, value in backup.items():
                setattr(self, key, value)


def from_dict(data_class_type, data):
    if data is None:
        return None
    if hasattr(data_class_type, '__dataclass_fields__'):  # It's a DataClass
        field_types = dict(data_class_type.__annotations__.items())
        return data_class_type(**{f: from_dict(field_types[f], data[f]) for f in data})
    if hasattr(data_class_type, '__origin__'):  # It's a generic
        if data_class_type.__origin__ is list:  # It's a List
            return [from_dict(data_class_type.__args__[0], d) for d in data]
        if data_class_type.__origin__ is dict:  # It's a Dict
            return {k: from_dict(data_class_type.__args__[1], v) for k, v in data.items()}
        raise NotImplementedError(f'Unknown generic type: {data_class_type}')
    # It's a regular type
    return data_class_type(data)


def is_stopping() -> bool:
    return app.is_stopping


def is_test() -> bool:
    return 'pytest' in sys.modules


class PackagePathFilter(logging.Filter):
    ABS_SYS_PATHS = sorted((os.path.abspath(p) + os.sep for p in sys.path), key=len, reverse=True)

    # https://stackoverflow.com/a/52582536/3419103
    @functools.lru_cache(maxsize=100)
    def find_relative_path(self, pathname: str) -> Optional[str]:
        for path in self.ABS_SYS_PATHS:
            if pathname.startswith(path):
                return os.path.relpath(pathname, path)
        return None

    def filter(self, record: logging.LogRecord) -> bool:
        record.relative_path = self.find_relative_path(record.pathname)
        return True



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import Sequence

from typing_extensions import Self


@dataclass(slots=True, kw_only=True)
class TimeBucket:
    date: str
    incidents: dict[str, int] = field(default_factory=dict)

    @classmethod
    def from_buckets(cls, sources: Sequence[TimeBucket]) -> Self:
        counter: Counter = Counter()
        for source in sources:
            counter.update(source.incidents)
        return cls(date=sources[0].date, incidents=dict(counter))


@dataclass(slots=True, kw_only=True)
class Day(TimeBucket):
    pass


@dataclass(slots=True, kw_only=True)
class Week(TimeBucket):
    pass


@dataclass(slots=True, kw_only=True)
class Month(TimeBucket):
    pass



from .module import ModuleHardware
from .robot_brain import RobotBrain


class BluetoothHardware(ModuleHardware):

    def __init__(self, robot_brain: RobotBrain, *, name: str = 'robot') -> None:
        self.name = name
        lizard_code = f'bluetooth = Bluetooth("{name}")'
        super().__init__(robot_brain, lizard_code=lizard_code)



import asyncio
import logging
import struct

from typing_extensions import Self


class AsyncTcpClient:
    """Async TCP client wrapper for sending and receiving messages to/from a server."""

    def __init__(self, ip: str, port: int) -> None:
        self.ip = ip
        self.port = port
        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None

    async def __aenter__(self) -> Self:
        self.reader, self.writer = await asyncio.open_connection(self.ip, self.port)
        return self

    async def __aexit__(self, *_) -> None:
        assert self.writer is not None
        self.writer.close()
        await self.writer.wait_closed()

    async def read(self, n_bytes: int, timeout_s: float = 3) -> bytes:
        """Read n_bytes from the server with a timeout.

        :param n_bytes: Number of bytes to read
        :param timeout_s: Timeout in seconds
        :return: Bytes read from the server

        :raises TimeoutError: If no response is received within the timeout
        :raises RuntimeError: If the response length is not equal to n_bytes
        """
        assert self.reader is not None
        data = await asyncio.wait_for(self.reader.read(n_bytes), timeout_s)
        if not data:
            raise TimeoutError('No response received within timeout')
        if not len(data) == n_bytes:
            raise RuntimeError(f'Received response with unexpected length {len(data)}')

        return data

    async def write(self, message: bytes) -> None:
        """Write a message to the server.

        :param message: Message to send
        """
        assert self.writer is not None
        self.writer.write(message)


class MotecSettingsInterface:

    def __init__(self, ip: str, port: int) -> None:
        self.ip = ip
        self.port = port

        self.log = logging.getLogger('MotecSettingsInterface')

        self.cooldown = 0.2
        self.cooldown_time = 0.0

        self.event_loop = asyncio.get_event_loop()

    def _construct_message(self, command_id: int, value_id: int, values: list[int]) -> bytes:
        n_value_bytes = len(values)
        if n_value_bytes > 6:
            raise ValueError('Too many values provided')

        v = [0] * 6
        for i, value in enumerate(values):
            v[i] = value

        return struct.pack('B' * 12, 48, 2, command_id, n_value_bytes + 1, value_id, 0, *v)

    async def _get_value(self, value_id: int, timeout_s: float = 3) -> bytearray:
        message = self._construct_message(70, value_id, [])
        self.log.debug('Sending message [get]: %s', message)

        # send message
        async with AsyncTcpClient(ip=self.ip, port=self.port) as client:
            await client.write(message)

            n_receive_bytes = 12

            # receive response asynchronously with timeout
            data = await client.read(n_receive_bytes, timeout_s)

        values = struct.unpack('B' * n_receive_bytes, data)[6:]
        self.log.debug('Received message [get]: %s', values)

        self.cooldown_time = self.event_loop.time() + self.cooldown

        return bytearray(values)

    async def _set_value(self, value_id: int, values: list[int] | int) -> None:
        if isinstance(values, int):
            values = [values]

        message = self._construct_message(38, value_id, values)
        self.log.debug('Sending message [set]: %s', message)
        # send message
        async with AsyncTcpClient(ip=self.ip, port=self.port) as client:
            await client.write(message)

        self.cooldown_time = asyncio.get_event_loop().time() + self.cooldown

    async def get_fps(self) -> int:
        return (await self._get_value(178))[0]

    async def set_fps(self, fps: int) -> None:
        await self._set_value(178, fps)

    async def get_stream_compression(self) -> int:
        return (await self._get_value(177))[0]

    async def set_stream_compression(self, level: int) -> None:
        if not 1 <= level <= 4:
            raise ValueError('Compression level must be between 1 and 4')

        await self._set_value(177, level)

    async def get_stream_resolution(self) -> tuple[int, int]:
        wmsb, wlsb, hmsb, hlsb = (await self._get_value(179))[:4]
        return (wmsb << 8) + wlsb, (hmsb << 8) + hlsb

    async def set_stream_resolution(self, width: int, height: int) -> None:
        wmsb = width >> 8
        wlsb = width & 0xFF
        hmsb = height >> 8
        hlsb = height & 0xFF
        await self._set_value(179, [wmsb, wlsb, hmsb, hlsb])

    async def get_stream_port(self) -> int:
        pmsb, plsb = (await self._get_value(200))[:2]
        return (pmsb << 8) + plsb

    async def set_stream_port(self, port: int) -> None:
        pmsb = port >> 8
        plsb = port & 0xFF
        await self._set_value(200, [pmsb, plsb])
